package MLocarna;
    
use strict;
use warnings;

BEGIN {
    use Exporter   ();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    
    # set the version for version checking
    $VERSION     = 1.00;
    # if using RCS/CVS, this may be preferred
    $VERSION = sprintf "%d.%03d", q$Revision: 1.3 $ =~ /(\d+)/g;
    
    @ISA         = qw(Exporter);
    @EXPORT      = qw(
		      &parse_mfasta
		      &parse_mfasta_constraints
		      &compute_alignment_from_seqs
		      &compute_alignment_from_dps
		      &compute_alignment_score
		      &parse_alignment
		      &parse_pp_file_aln
		      &parse_pp_file_pairprobs
		      &convert_dp_to_pp_with_constraints
		      );
    %EXPORT_TAGS = ( );     # eg: TAG => [ qw!name1 name2! ],
    
    # your exported package globals go here,
    # as well as any optionally exported functions
    @EXPORT_OK   = qw();
}
our @EXPORT_OK;


#
# read a (multiple) fasta file and return
# a hash that maps the sequence names to their
# corresponding sequences
#
sub parse_mfasta($) {
    my ($file) = @_;
    my %mfasta;
    
    local *PMF_IN;
    
    open(PMF_IN,$file) || die "Cannot read mfasta file $file\n";

    my $line=<PMF_IN>;
    while(defined($line)) {
	if ($line=~/^>\s*(.+)/) {
	    my $name=$1;
	    chomp $name;
	    $name =~ s/[^a-zA-Z\d]/_/g;
	    $name = substr $name,0,16;
	    
	    if (exists $mfasta{$name}) {
		print STDERR "Duplicate name in mfasta: $name\n";
		exit -1;
	    }
	    while (defined($line=<PMF_IN>) && ($line !~ /^>/)) {
		chomp $line;
		$line =~ s/\s+//g;
		$mfasta{$name} .= $line;
	    }
	}
    }
    return %mfasta;
}

#
# generate a sequence name from $name that
# has at most a length of 16
# and
# does not already exist in @names
# 
#
sub make_unique_seqname {
    my ($name, @names) = @_;
    
    my $maxlen=16;

    chomp $name;
    $name =~ s/[^a-zA-Z\d]/_/g;
    $name = substr $name,0,$maxlen;
    
    my $i=1;
    while (grep /^$name$/, @names) {
	my $arity = int(log($i)/log(10))+1;
	if ($arity >= 4) { # arbitrary limit
	    # give up!
	    last;
	}
	
	$name = substr $name,0,$maxlen-$arity-1;
	$name = sprintf("%s_%0$arity"."d",$name,$i);
	$i++;
    }
    return $name;
}


#
# read a (multiple) fasta file 
# with constraint annotation
# also reads long names (the complete name in the fasta file)
#
# constraints are specified as strings of the same length
# as the sequence. The idea is:
#
#  - a string followed by #S on the same line defines
#    structure constraints (this is passed to RNAfold -p!)
#  - a string followed by #<k> is the k-th line of
#    the sequence constraint (this is passed to locarna)
#
# lines of different "types" can be mixed freely!
#
# return a hash that maps the sequence names to their
# corresponding sequences and constraints:
#
#  for a sequence name $name, we generate hash entries
#    - $mfasta{"$name"} 
#    - $mfasta{"$name#S"} 
#    - $mfasta{"$name#C"}
#    - $mfasta{"$name#LONG"}
#
sub parse_mfasta_constraints {
    my ($file) = @_;

    my %mfasta;

    my %seqcons; # sequence constraints hash
    my %strcons; # sequence constraints hash
    
    local *PMF_IN;
    
    open(PMF_IN,$file) || die "Cannot read mfasta file $file\n";

    my $line=<PMF_IN>;
    while(defined($line)) {
	if ($line=~/^>\s*(.+)/) {
	    
	    my $longname = $1;
	    $longname =~ s/\s*$//; # eat space at the end of $longname

	    my $name = make_unique_seqname $longname,(keys %mfasta);
	    
	    if (exists $mfasta{$name}) {
		print STDERR "Duplicate name in mfasta, cannot make unique : $name\n";
		exit -1;
	    }
	    
	    $mfasta{"$name\#LONG"}=$longname;
	    
	    while (defined($line=<PMF_IN>) && ($line !~ /^>/)) {
		chomp $line;
		$line =~ s/\s+//g;
		if ($line =~ /(.+)\#S/) {
		    $strcons{$name} .= $1;
		} elsif ($line =~ /(.+)\#(.+)/) {
		    $seqcons{"$name\#$2"} .= $1;
		} else {
		    $mfasta{$name} .= $line;
		}
	    }
	}
    }
    
    my @names = keys %mfasta;
    
    ## ----------------------------------------
    # insert constraints in mfasta hash
    # and check validity
    #
    my $has_seq_constraints=0; # whether there are seq constraints
    
    for my $name (@names) {
	if (exists $strcons{$name}) {
	    if (length $strcons{$name} != length $mfasta{$name}) {
		print STDERR "Structure constraint length unequals sequence length for $name.\n";
		exit -1;
	    }

	    $mfasta{"$name\#S"} = $strcons{$name};
	}
	
	my $i=1;
	while (exists $seqcons{"$name\#$i"}) {
	    if (length $seqcons{"$name\#$i"} != length $mfasta{$name}) {
		print STDERR "Sequence constraint length unequals sequence length for $name.\n";
		exit -1;
	    }
	    
	    $mfasta{"$name\#C"} .= $seqcons{"$name\#$i"}."\#";
	    $i++;
	}
	if ($i>1) {
	    $mfasta{"$name\#C"} =~ s/\#$//; # remove last '#'
	    $has_seq_constraints=1;
	}
	
	my $num_seq_constraint_lines=grep /$name\#/, keys %seqcons;
	if ($i-1 != $num_seq_constraint_lines) {
	    print STDERR "Bad sequence constraints for sequence $name.\n";
	    exit -1;
	}
    }
    
    for my $name (@names) {
	if (($name!~/#/) &&  $has_seq_constraints) {
	    if (! exists $mfasta{"$name\#C"}) {
		print STDERR "Sequence constraint string missing for $name.\n";
	    }
	}
    }
    #
    ## ----------------------------------------
    
    return %mfasta;
}

## convert a dotplot file to a pp file
## thereby insert the sequence constraint strings
sub convert_dp_to_pp_with_constraints($$$$$) {
    my ($dpfile,$ppfile,$name,$sequence,$constraints) = @_;
    
    open(PP_OUT,">$ppfile") || die "Cannot open $ppfile for writing."; 
    open(DP_IN,"$dpfile") || die "Cannot open $dpfile for reading."; 
    
    print PP_OUT "SCORE: 0\n\n";
    print PP_OUT "$name $sequence\n";
    if (defined $constraints) {
	print PP_OUT "#C $constraints\n";
    }
    
    my %pp;
    
    while (my $line=<DP_IN>) {
	if ($line =~ /^(\d+) (\d+) ([\d\.]+) ubox/) {
	    my $i=$1;
	    my $j=$2;
	    my $p=$3*$3; 
	    $pp{$i}{$j}="$p";
	}
	
	# assume ubox always before lbox
	if ($line =~ /^(\d+) (\d+) ([\d\.]+) lbox/) {
	    my $i=$1;
	    my $j=$2;
	    my $p=$3*$3; 
	    if (! exists $pp{$i}{$j}) {
		$pp{$i}{$j}="0";
	    }
	    $pp{$i}{$j}.=" $p";
	}
    }
    
    close DP_IN;
    
    print PP_OUT "\n\#\n";
    
    for my $i ( keys %pp ) {
	for my $j ( keys %{ $pp{$i} } ) {
	    print PP_OUT "$i $j $pp{$i}{$j}\n";
	}
    }
    
    close PP_OUT;
}


## compute pairwise alignment for given sequences using RNAfold -p for generating dps
sub compute_alignment($$$$$) {
    my ($bindir,$seqA,$seqB,$locarna_params,$tmpprefix) =  @_;
    
    local *CA_IN;
    
    my $tmpfile="$tmpprefix.clustal";

    system("$bindir/locarna.pl $seqA $seqB $locarna_params --clustal=$tmpfile >/dev/null");
    
    open(CA_IN,"$tmpfile") || die "Cannot read tmp file $tmpfile\n";
    
    my @content=<CA_IN>;
   
    close CA_IN;
    
    unlink $tmpfile;
    
    return @content;
}


## compute the alignment from dp-files or pp-files
sub compute_alignment_from_dps($$$$$) {
    my ($bindir,$seqA_dp,$seqB_dp,$locarna_params,$tmpprefix) =  @_;

    local *CA_IN;

    my $tmpfile="$tmpprefix.clustal";

    my $cmd="$bindir/locarna $seqA_dp $seqB_dp $locarna_params --clustal=$tmpfile >/dev/null";
    system $cmd || die "Command $cmd failed."; 
    
    open(CA_IN,"$tmpfile") || die "compute_alignment: Cannot read temporary file $tmpfile\n";
    
    my @content=<CA_IN>;
   
    close CA_IN;
    
    unlink $tmpfile;
    
    return @content;
}


sub compute_alignment_score($$$$$) {
    my ($bindir,$seqA,$seqB,$locarna_params,$tmpprefix) =  @_;
    
    my @content=compute_alignment($bindir,$seqA,$seqB,$locarna_params,$tmpprefix);
    # print "@content\n";
    $content[0] =~ /Score: ([\d\.\-]+)/ || die "Cannot return score.\n";
    my $score=$1;

    return $score;
}


sub parse_alignment {
    my ($nameA,$nameB,$alignment_ref) = @_;

    my @alignment = @{ $alignment_ref };
    
    my $aliA="";
    my $aliB="";
    
    for my $line (@alignment) {
	if ($line =~ /^$nameA\s+(.+)/) {
	    $aliA.=$1;
	} elsif ($line =~ /^$nameB\s+(.+)/) {
	    $aliB.=$1;
	}
    }
    return ($aliA,$aliB);
}


sub parse_pp_file_aln($) {
    my ($filename)=@_;
    local *PP_IN;
   
    open(PP_IN,$filename) || die "Can not read $filename\n";
    
    my %aln;
    my %pairprobs;

    my $line;
    
    while ($line = <PP_IN>) {
	if ($line =~ /^\#\s*$/) { last; }
    
	if (($line =~ /^(\S+)\s+(.+)/) && ($line !~ /^SCORE:/) ) {
	    $aln{$1}.=$2;
	}
    }
    
    close PP_IN;
    
    return %aln;
}

sub parse_pp_file_pairprobs($) {
    my ($filename)=@_;
    local *PP_IN;

    open(PP_IN,$filename) || die "Can not read $filename\n";
    
    my %aln;
    my %pairprobs;

    my $line;

    while ($line = <PP_IN>) {
	if ($line =~ /^\#\s*$/) { last; }

	if ($line =~ /^(\S+)\s+(.+)/) {
	    $aln{$1}.=$2;
	}
    }
    
    while (($line = <PP_IN>)) {
	if ($line =~ /(\S+)\s+(\S+)\s+(\S+)/) {
	    my $i=$1;
	    my $j=$2;
	    my $p=$3;
	    $pairprobs{"$i $j"}=$p;
	}
    }
    close PP_IN;
    
    return %pairprobs;
}




## ------------------------------------------------------------

END {}

1;
