#ifndef ALIGNER_P_HH
#define ALIGNER_P_HH

#include "sequence.hh"
#include "basepairs.hh"

#include "params.hh"
#include "scoring.hh"

#include "matrices.hh"

#include "arc_matches.hh"

/**
   Contains data for restricting AlignerP to sub-sequences
*/
class AlignerPRestriction {
  typedef size_t size_type;
private:    
    size_type startA;
    size_type startB;
    size_type endA;
    size_type endB;
public:
    AlignerPRestriction(size_type startA_, size_type startB_, size_type endA_, size_type endB_)
	: startA(startA_), startB(startB_), endA(endA_), endB(endB_)
    {}

    size_t get_startA() const {//std::cout<<"StartA "<<startA<<std::endl;
    return startA;}
    size_t get_endA() const {//std::cout<<"endA "<<endA<<std::endl; 
    	return endA;}
    
    size_t get_startB() const {//std::cout<<"StartB "<<startB<<std::endl;
    return startB;}
    size_t get_endB() const {//std::cout<<"endB "<<endB<<std::endl; 
    	return endB;}

    void set_startA(size_t p) {startA=p;}
    void set_endA(size_t p)  {endA=p;}
    void set_startB(size_t p)  {startB=p;}
    void set_endB(size_t p)  {endB=p;}
};

inline
std::ostream & operator<<(std::ostream &out, AlignerPRestriction r) {
    return
	out << r.get_startA() << " "
	    << r.get_startB() << " "
	    << r.get_endA() << " "
	    << r.get_endB();
}

/* ============================================================ 
   class AlignerP

   Compute partition function of alignment, arc match and base match probabilities
   
   initialize with two sequences and their basepairs (including scores)
   ============================================================ 
*/

#define FOR_IJ_IN_ARCMATCH \
  for (size_type i=al+1; i<ar; i++) \
    for (size_type j=bl+1; j<br; j++)

//! class for doing the alignment of two sequences and their
//! associated sets of weighted basepairs
//!
//! an object always knows about the two sequences and
//! the two weighted base pair sets
//!
//! common usage: construct, align, trace, get_alignment
class AlignerP {
public:
    typedef size_t size_type;
protected:
    const Scoring *scoring; //!< the scores
    
    const AlignerParams *params; //!< the parameter for the alignment
	
    const Sequence &seqA;
    const BasePairs &bpsA;
    const Sequence &seqB;
    const BasePairs &bpsB;
    
    const ArcMatches &arc_matches;


    // AlignerP always works on sub-sequences/structures of
    // seqA and seqB. These are specified by
    AlignerPRestriction r;
    // in the standard case, i.e. align the whole sequences,
    // the values are set to startA=1 and endA=seqA.length()
    // and analogously for B
	
    // the constructor initializes to the above standard values
    // methods provide possibility to restriction

    

    pf_score_t partFunc; //!< the total partition function (only defined after call of align_inside())

    /**
       D(a,b) is the partition function of the subsequences seqA(al..ar) and seqB(bl..br),
       where the arcs a and b match
    */
    PFScoreMatrix Dmat;
    

    /**
       For the current pair of left arc ends (al,bl) and a current line i
       E(j) is the partition function of the subsequences seqA(al+1..i) and seqB(bl+1..j)
       covering only alignments that gap the last position of seqA
       
       In the algorithm, this is constantly overwritten, i.e. for a current j
       all entries E(j') j'<j are for the current line i and all entries j'>j are for the
       line i-1
    */
    PFScoreVector E;

    /**
       For the current pair of left arc ends (al,bl) and current indices (i,j),
       F is the the partition function of the subsequences seqA(al+1..i) and seqB(bl+1..j)
       covering only alignments that gap the last position of seqB
       
       In the algorithm, this is constantly overwritten, i.e. when we compute the entries for (i,j)
       it will still contain the value of (i,j-1) and is then updated to the value for (i,j)
    */
    pf_score_t F;
    
    
    /**
       For the current pair of left arc ends (al,bl),
       M(i,j) is the partition function of the subsequences seqA(al+1..i) and seqB(bl+1..j)
    */
    PFScoreMatrix M;


    /**
       For the current pair of left arc ends (al,bl),
       M(i,j) is the partition function of the subsequences seqA(i+1..al-1) and seqB(j+1..bl-1)
    */
    PFScoreMatrix Mrev;
    
    PFScoreVector Erev;
    pf_score_t    Frev;
    
    /**
       D'(a,b) is the partition function of the subsequences seqA(1..al-1,ar+1..lenA) and seqB(1..bl-1,br+1..lenB)
       times the score contribution of the arc match (al,ar);(bl,br)
    */
    PFScoreMatrix Dmatprime;

    /**
      For the current pair of left arc ends (al,bl) and line i,
      E'(j) is the partition function of the subsequences seqA(1..al-1,i+1..lenA) and seqB(1..bl-1,j+1..lenB)
      where i+1 is aligned to a gap
     */
    PFScoreVector Eprime; // one could slightly optimize space by PFScoreVector &Eprime = E;
    
    /**
      For the current pair of left arc ends (al,bl) and (i,j),
      F' is the partition function of the subsequences seqA(1..al-1,i+1..lenA) and seqB(1..bl-1,j+1..lenB)
      where j+1 is aligned to a gap
     */
    pf_score_t Fprime;

    /**
       For the current pair of left arc ends (al,bl),
       M'(i,j) is the partition function of the subsequences seqA(1..al-1,i+1..lenA) and seqB(1..bl-1,j+1..lenB)
    */
    PFScoreMatrix Mprime;

    
    

    //! probabilities of arc matchs, as computed by the algo
    ProbMatrix am_prob;

    //! probabilities of base matchs, as computed by the algo
    ProbMatrix bm_prob;
    
    bool D_created; //!< flag, is D already created?
    bool Dprime_created; //!< flag, is Dprime already created?


    bool am_heuristic;

    //! initialize first column and row of M, for inside recursion
    void init_M(size_type al, size_type ar, size_type bl, size_type br);

    //! initialize E
    void init_E(size_type al, size_type ar, size_type bl, size_type br);
    
    //! initialize the reversed M matrix
    void init_Mrev(size_type al, size_type ar, size_type bl, size_type br); 

    //! initialize the reversed E matrix/vector
    void init_Erev(size_type al, size_type ar, size_type bl, size_type br); 

    //! initialize first column and row of M' for outside recursion
    void init_Mprime(size_type al, size_type ar, size_type bl, size_type br);

    //! initialize first row of E' for outside recursion
    void init_Eprime(size_type al, size_type ar, size_type bl, size_type br);

    //! compute one entry in E (inside recursion cases)
    pf_score_t comp_E_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in F (inside recursion cases)
    pf_score_t comp_F_entry(size_type al, size_type bl, size_type i, size_type j);
    
    //! compute one entry in M (inside recursion cases)
    pf_score_t comp_M_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in Mprime (outside recursion cases)
    pf_score_t comp_Mprime_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in Eprime (outside recursion cases)
    pf_score_t comp_Eprime_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in Fprime (outside recursion cases)
    pf_score_t comp_Fprime_entry(size_type al, size_type bl, size_type i, size_type j);


    //! compute one entry in Erev
    pf_score_t comp_Erev_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in Frev
    pf_score_t comp_Frev_entry(size_type al, size_type bl, size_type i, size_type j);

    //! compute one entry in Mrev
    pf_score_t comp_Mrev_entry(size_type al, size_type bl, size_type i, size_type j);
    
    //! align, where al,ar,bl,br is an arc-match
    void align_inside_arcmatch(size_type al,size_type ar,size_type bl,size_type br);
    
    //! align, where al,ar,bl,br is an arc-match
    void align_outside_arcmatch(size_type al,size_type ar,size_type bl,size_type br);
    
    //! align, reversed
    void align_reverse(size_type al, size_type bl,size_type i, size_type j);
    
    //! create the entries in the D matrix.
    //! This function is called by align() (unless D_created)
    void align_D();
    
    //! create the entries in the Dprime matrix
    //!   This function is called by align() (unless Dprime_created)
    //! uses inside recursion
    void align_Dprime();
    
    //!  fill in D the entries with left ends al,bl, 
    //! where adjlA, adjlB are adjanceny lists of the arcs
    //! uses inside recursion
    void 
    fill_D(size_type al, size_type bl,
	   size_type max_ar, size_type max_br);
    
    //!  fill in D the entries with right ends ar,br, 
    //! where adjrA, adjrB are adjanceny lists of the arcs
    //! uses outside recursion
    void 
    fill_Dprime(size_type al, size_type bl,
		size_type min_ar, size_type min_br
		);


    pf_score_t &D(const ArcMatch &am) {
	return Dmat(am.arcA().idx(),am.arcB().idx());
    }

    pf_score_t &D(const Arc &arcA,const Arc &arcB) {
	return Dmat(arcA.idx(),arcB.idx());
    }

    pf_score_t &Dprime(const ArcMatch &am) {
	return Dmatprime(am.arcA().idx(),am.arcB().idx());
    }

    pf_score_t &Dprime(const Arc &arcA,const Arc &arcB) {
	return Dmatprime(arcA.idx(),arcB.idx());
    }
    
public:  
    
    //! construct with sequences and corresponding sets of basepairs
    AlignerP(const Sequence &seqA,
	     const Sequence &seqB,
	     const ArcMatches &arc_matches, 
	     const AlignerParams *ap,
	     const Scoring *s
	    );
    
    //! compute the partition function by the inside algorithm
    //! and fill the D matrix
    pf_score_t align_inside();
       
    //! perform the outside algorithm
    //! and fill the Dprime matrix,
    //! assumes that D matrix is computed already
    void align_outside();
    
    //! computes the probabilitites of all base matches and stores them internally (in a matrix)
    void compute_basematch_probabilities( bool basematch_probs_include_arcmatch );

    //! computes the probabilitites of all arc matches and stores them internally (in a sparse matrix)
    void compute_arcmatch_probabilities();
  
    void write_arcmatch_probabilities(std::ostream &out);
    void write_basematch_probabilities(std::ostream &out); 
    
    
    // const Matrix *get_basematch_probabilities(){}
    // const SparseMatrix *get_arcmatch_probabilities(){}    
};

#endif
